@@ -1,4 +1,3 @@
-import functools
 import warnings
 
 
@@ -10,19 +9,15 @@ def warn_experimental(msg):
     warnings.warn(msg, category=ExperimentalWarning, stacklevel=2)
 
 
-def mark_experimental(msg=None):
+def mark_experimental(f, msg=None):
     def inner(f):
         message = msg
         if message is None:
-            full_name = f"{f.__module__}.{f.__name__}"
+            full_name = f"{f.__module__}.{f.__class__.__name__}"
             message = f"Function {full_name} is experimental and may cause any kind of troubles"
 
-        @functools.wraps(f)
-        def wrapped(*args, **kwargs):
-            warn_experimental(message)
-            return f(*args, **kwargs)
-
-        return wrapped
+        warn_experimental(message)
+        return f
 
     return inner
 
