@@ -1,5 +1,13 @@
+import dataclasses
+from functools import lru_cache, wraps
+from itertools import count
+from pathlib import Path
+from typing import Generator, Iterator, List, Optional, Union
+from joblib import Parallel, delayed
+
 import feets
 import numpy as np
+import pandas as pd
 import pytest
 from numpy.testing import assert_allclose
 from scipy import stats
@@ -9,6 +17,88 @@ import light_curve.light_curve_ext as lc_ext
 import light_curve.light_curve_py as lc_py
 
 
+@dataclasses.dataclass
+class Data:
+    name: str
+    phot_type: str
+    t: np.ndarray = dataclasses.field(repr=False)
+    m: np.ndarray = dataclasses.field(repr=False)
+    sigma: np.ndarray = dataclasses.field(repr=False)
+
+    # Kinda static attribute
+    phot_type_choices: frozenset = dataclasses.field(
+        default=frozenset(["flux", "mag"]),
+        init=False,
+        repr=False,
+    )
+
+    # For '*' operator
+    def __iter__(self):
+        def gen():
+            yield self.t
+            yield self.m
+            yield self.sigma
+
+        return gen()
+
+    def __post_init__(self):
+        self.name = str(self.name)
+        assert self.phot_type in self.phot_type_choices
+        assert self.t.size == self.m.size == self.sigma.size
+        assert self.t.dtype == self.m.dtype == self.sigma.dtype
+
+
+def gen_data_from_test_data_path(
+    paths: Iterator[Union[Path, str]], *, n: Optional[int] = None, convert_to_flux: bool = False
+) -> Generator[Data, None, None]:
+    if n is None:
+        take_n = count()
+    else:
+        take_n = range(n)
+    for _, csv_file in zip(take_n, paths):
+        df = pd.read_csv(csv_file)
+
+        # Drop repeated values
+        idx = np.diff(df["time"], prepend=0) > 0
+        df = df[idx]
+
+        t = df["time"].to_numpy()
+        if "mag" in df.columns:
+            m = df["mag"].to_numpy()
+            sigma = df["magerr"].to_numpy()
+            if convert_to_flux:
+                m = 10 ** (-0.4 * m)
+                sigma = 0.4 * np.log(10.0) * sigma * m
+                phot_type = "flux"
+            else:
+                phot_type = "mag"
+        elif "flux" in df.columns:
+            m = df["flux"].to_numpy()
+            sigma = df["fluxerr"].to_numpy()
+            phot_type = "flux"
+        else:
+            raise ValueError(f"No mag neither flux column in {csv_file}")
+        yield Data(csv_file, phot_type, t, m, sigma)
+
+
+@lru_cache(maxsize=1)
+def get_test_data_from_issues() -> List[Data]:
+    data_root = Path(__file__).parent / "light-curve-test-data/from-issues"
+    return list(gen_data_from_test_data_path(data_root.glob("**/*.csv")))
+
+
+@lru_cache()
+def get_first_n_snia_data(*, n: Optional[int] = None, convert_to_flux: bool = False) -> List[Data]:
+    data_root = Path(__file__).parent / "light-curve-test-data/SNIa"
+    with open(data_root / "snIa_bandg_minobs10_beforepeak3_afterpeak4.csv") as fh:
+        file_names = frozenset(fh.read().split())
+    light_curve_dir = data_root / "light-curves"
+    # Sort to make sure the order is always the same
+    csv_paths = sorted(light_curve_dir.glob("*.csv"))
+    paths = (path for path in csv_paths if path.stem in file_names)
+    return list(gen_data_from_test_data_path(paths, n=n, convert_to_flux=convert_to_flux))
+
+
 class _Test:
     # Feature name must be updated in child classes
     name = None
@@ -26,6 +116,9 @@ class _Test:
     # Specify a `str` with a reason why skip this test
     feets_skip_test = False
 
+    # Which types of pre-generated light-curves to use
+    phot_types = Data.phot_type_choices
+
     def setup_method(self):
         self.rust = getattr(lc_ext, self.name)(*self.args)
         try:
@@ -37,7 +130,7 @@ class _Test:
             self.feets_extractor = feets.FeatureSpace(only=[self.feets_feature], data=["time", "magnitude", "error"])
 
     # Default values of `assert_allclose`
-    rtol = 1e-7
+    rtol = np.finfo(np.float32).resolution  # 1e-6
     atol = 0
 
     # Default values for random light curve generation
@@ -51,72 +144,127 @@ class _Test:
 
     add_to_all_features = True
 
-    def generate_data(self):
+    def random_data(self):
         t = np.sort(np.random.uniform(self.t_min, self.t_max, self.n_obs))
         m = np.random.uniform(self.m_min, self.m_max, self.n_obs)
         sigma = np.random.uniform(self.sigma_min, self.sigma_max, self.n_obs)
-        return t, m, sigma
+        return Data("random", "mag", t, m, sigma)
 
-    def test_feature_length(self):
-        t, m, sigma = self.generate_data()
-        result = self.rust(t, m, sigma, sorted=None)
-        assert len(result) == len(self.rust.names) == len(self.rust.descriptions)
+    def from_issues_data(self):
+        for data in get_test_data_from_issues():
+            if data.phot_type not in self.phot_types:
+                continue
+            yield data
 
-    def test_close_to_lc_py(self):
+    def snia_data(self, n: Optional[int] = None):
+        for data in get_first_n_snia_data(n=n):
+            if data.phot_type not in self.phot_types:
+                continue
+            yield data
+
+    def data_gen(self) -> Generator[Data, None, None]:
+        yield self.random_data()
+        yield from self.from_issues_data()
+        yield from self.snia_data(n=10)
+
+    def test_feature_length(self, subtests):
+        for data in self.data_gen():
+            with subtests.test(lc_name=data):
+                result = self.rust(*data, sorted=None)
+                assert len(result) == len(self.rust.names) == len(self.rust.descriptions)
+
+    def test_close_to_lc_py(self, subtests):
         if self.py_feature is None:
             pytest.skip("No matched light_curve_py class for the feature")
-        t, m, sigma = self.generate_data()
-        assert_allclose(self.rust(t, m, sigma), self.py_feature(t, m, sigma), rtol=self.rtol, atol=self.atol)
+        for data in self.data_gen():
+            with subtests.test(data=data):
+                assert_allclose(self.rust(*data), self.py_feature(*data), rtol=self.rtol, atol=self.atol)
 
     def test_benchmark_rust(self, benchmark):
-        t, m, sigma = self.generate_data()
-
+        t, m, sigma = self.random_data()
         benchmark.group = str(type(self).__name__)
         benchmark(self.rust, t, m, sigma, sorted=True, check=False)
 
+    @pytest.mark.nobs
+    @pytest.mark.parametrize("n_obs", np.logspace(1, 6, 6))
+    def test_benchmark_rust_n(self, benchmark, n_obs):
+        self.n_obs = int(n_obs)
+        t, m, sigma = self.random_data()
+
+        benchmark.group = f"{n_obs}_observations"
+        benchmark.name = f"{str(type(self).__name__)}_rust"
+        benchmark(self.rust, t, m, sigma, sorted=True, check=False)
+
     def test_benchmark_lc_py(self, benchmark):
         if self.py_feature is None:
             pytest.skip("No matched light_curve_py class for the feature")
 
-        t, m, sigma = self.generate_data()
+        t, m, sigma = self.random_data()
 
         benchmark.group = str(type(self).__name__)
         benchmark(self.py_feature, t, m, sigma, sorted=True, check=False)
 
-    def test_close_to_naive(self):
+    @pytest.mark.nobs
+    @pytest.mark.parametrize("n_obs", np.logspace(1, 6, 6))
+    def test_benchmark_lc_py_n(self, benchmark, n_obs):
+        if self.py_feature is None:
+            pytest.skip("No matched light_curve_py class for the feature")
+
+        self.n_obs = int(n_obs)
+        t, m, sigma = self.random_data()
+
+        benchmark.group = f"{n_obs}_observations"
+        benchmark.name = f"{str(type(self).__name__)}_py"
+        benchmark(self.py_feature, t, m, sigma, sorted=True, check=False)
+
+    def test_close_to_naive(self, subtests):
         if self.naive is None:
             pytest.skip("No naive implementation for the feature")
-
-        t, m, sigma = self.generate_data()
-        assert_allclose(self.rust(t, m, sigma), self.naive(t, m, sigma), rtol=self.rtol, atol=self.atol)
+        for data in self.data_gen():
+            with subtests.test(data=data):
+                assert_allclose(self.rust(*data), self.naive(*data), rtol=self.rtol, atol=self.atol)
 
     def test_benchmark_naive(self, benchmark):
         if self.naive is None:
             pytest.skip("No naive implementation for the feature")
 
-        t, m, sigma = self.generate_data()
+        t, m, sigma = self.random_data()
 
         benchmark.group = type(self).__name__
         benchmark(self.naive, t, m, sigma)
 
+    @pytest.mark.nobs
+    @pytest.mark.parametrize("n_obs", np.logspace(1, 6, 6))
+    def test_benchmark_naive_n(self, benchmark, n_obs):
+        if self.naive is None:
+            pytest.skip("No naive implementation for the feature")
+
+        self.n_obs = int(n_obs)
+        t, m, sigma = self.random_data()
+
+        benchmark.group = f"{n_obs}_observations"
+        benchmark.name = f"{str(type(self).__name__)}_naive"
+        benchmark(self.naive, t, m, sigma)
+
     def feets(self, t, m, sigma):
         _, result = self.feets_extractor.extract(t, m, sigma)
         return result
 
-    def test_close_to_feets(self):
+    def test_close_to_feets(self, subtests):
         if self.feets_extractor is None:
             pytest.skip("No feets feature provided")
         if self.feets_skip_test:
             pytest.skip("feets is expected to be different from light_curve, reason: " + self.feets_skip_test)
 
-        t, m, sigma = self.generate_data()
-        assert_allclose(self.rust(t, m, sigma)[:1], self.feets(t, m, sigma)[:1], rtol=self.rtol, atol=self.atol)
+        for data in self.data_gen():
+            with subtests.test(data=data):
+                assert_allclose(self.rust(*data)[:1], self.feets(*data)[:1], rtol=self.rtol, atol=self.atol)
 
     def test_benchmark_feets(self, benchmark):
         if self.feets_extractor is None:
             pytest.skip("No feets feature provided")
 
-        t, m, sigma = self.generate_data()
+        t, m, sigma = self.random_data()
 
         benchmark.group = type(self).__name__
         benchmark(self.feets, t, m, sigma)
@@ -148,6 +296,8 @@ if lc_ext._built_with_gsl:
 
         add_to_all_features = False  # in All* random data is used
 
+        phot_types = frozenset(["flux"])
+
         @staticmethod
         def _model(t, a, b, t0, rise, fall):
             dt = t - t0
@@ -164,13 +314,19 @@ if lc_ext._built_with_gsl:
         # Random data yields to random results because target function has a lot of local minima
         # BTW, this test shouldn't use fixed random seed because the curve has good enough S/N to be fitted for any give
         # noise sample
-        def generate_data(self):
+        def random_data(self):
             rng = np.random.default_rng(0)
             t = np.linspace(self.t_min, self.t_max, self.n_obs)
             model = self._model(t, *self._params())
             sigma = np.sqrt(model)
             m = model + sigma * rng.normal(size=self.n_obs)
-            return t, m, sigma
+            return Data("Bazin+noise", "flux", t, m, sigma)
+
+        # Keep random data only
+        def data_gen(self) -> Generator[Data, None, None]:
+            yield self.random_data()
+            # All these fail =(
+            # yield from get_first_n_snia_data(n=10, convert_to_flux=True)
 
         def naive(self, t, m, sigma):
             params, _cov = curve_fit(
@@ -249,6 +405,28 @@ class TestInterPercentileRange(_Test):
     feets_skip_test = "feets uses different quantile type"
 
 
+class TestOtsuSplit(_Test):
+    name = "OtsuSplit"
+
+
+def magnitude_function(func, *args, **kwargs):
+    @wraps(func)
+    def decorated(t, m, sigma=None, sorted=None, check=None):
+        return func(m)
+
+    return decorated
+
+
+class TestOtsuSplitThreshold(_Test):
+    def setup_method(self):
+        self.py_feature = magnitude_function(lc_py.OtsuSplit.threshold)
+        self.rust = magnitude_function(lc_ext.OtsuSplit.threshold)
+
+    def test_feature_length(self):
+        """Not a real feature extractor, no need to test length"""
+        pass
+
+
 class TestKurtosis(_Test):
     name = "Kurtosis"
 
@@ -409,6 +587,9 @@ class TestWeightedMean(_Test):
 
 
 class TestAllPy(_Test):
+    # Most of the features are for mags
+    phot_types = frozenset(["mag"])
+
     def setup_method(self):
         features = []
         py_features = []
@@ -425,7 +606,104 @@ class TestAllPy(_Test):
         self.py_feature = lc_py.Extractor(*py_features)
 
 
+class TestBenchmarkParallel(_Test):
+    phot_types = frozenset(["mag"])
+    features_subset = [
+        "AndersonDarlingNormal",
+        "Beyond1Std",
+        "Cusum",
+        "EtaE",
+        "InterPercentileRange",
+        "Kurtosis",
+        "LinearTrend",
+        "MaximumSlope",
+        "Mean",
+        "MeanVariance",
+        "MedianAbsoluteDeviation",
+        "MedianBufferRangePercentage",
+        "PercentAmplitude",
+        "PercentDifferenceMagnitudePercentile",
+        "Skew",
+        "StandardDeviation",
+        "StetsonK",
+    ]
+
+    def setup_method(self):
+        features = []
+        py_features = []
+        for cls in _Test.__subclasses__():
+            if cls.name is None and cls.name not in self.features_subset:
+                continue
+
+            try:
+                py_features.append(getattr(lc_py, cls.name)(*cls.args))
+            except AttributeError:
+                continue
+            features.append(getattr(lc_ext, cls.name)(*cls.args))
+        self.rust = lc_ext.Extractor(*features)
+        self.py_feature = lc_py.Extractor(*py_features)
+
+    def rust_many(self, n_jobs, light_curves):
+        return self.rust.many(
+            light_curves,
+            n_jobs=int(n_jobs),
+            sorted=True,
+            check=False,
+        )
+
+    def py_parallel(self, n_jobs, light_curves):
+        return Parallel(n_jobs=int(n_jobs))(delayed(self.py_feature)(*lc) for lc in light_curves)
+
+    def rust_parallel(self, n_jobs, light_curves):
+        return Parallel(n_jobs=int(n_jobs))(delayed(self.rust)(*lc) for lc in light_curves)
+
+    @pytest.mark.parametrize("n_core", np.linspace(1, 8, 8))
+    @pytest.mark.multi
+    def test_benchmark_many_rust(self, benchmark, n_core):
+        light_curves = []
+
+        for _ in range(1000):
+            t, m, sigma = self.random_data()
+            light_curves.append((t, m, sigma))
+
+        benchmark.group = f"{n_core}_cores"
+        benchmark.name = "rust multiprocessing (many)"
+
+        benchmark(self.rust_many, n_core, light_curves)
+
+    @pytest.mark.parametrize("n_core", np.linspace(1, 8, 8))
+    @pytest.mark.multi
+    def test_benchmark_multi_python(self, benchmark, n_core):
+        light_curves = []
+
+        for _ in range(1000):
+            t, m, sigma = self.random_data()
+            light_curves.append((t, m, sigma))
+
+        benchmark.group = f"{n_core}_cores"
+        benchmark.name = "python multiprocessing"
+
+        benchmark(self.py_parallel, n_core, light_curves)
+
+    @pytest.mark.parametrize("n_core", np.linspace(1, 8, 8))
+    @pytest.mark.multi
+    def test_benchmark_multi_rust(self, benchmark, n_core):
+        light_curves = []
+
+        for _ in range(1000):
+            t, m, sigma = self.random_data()
+            light_curves.append((t, m, sigma))
+
+        benchmark.group = f"{n_core}_cores"
+        benchmark.name = "rust multiprocessing"
+
+        benchmark(self.rust_parallel, n_core, light_curves)
+
+
 class TestAllNaive(_Test):
+    # Most of the features are for mags
+    phot_types = frozenset(["mag"])
+
     def setup_method(self):
         features = []
         self.naive_features = []
@@ -447,6 +725,9 @@ class TestAllNaive(_Test):
 class TestAllFeets(_Test):
     feets_skip_test = "skip for TestAllFeets"
 
+    # Most of the features are for mags
+    phot_types = frozenset(["mag"])
+
     def setup_method(self):
         features = []
         feets_features = []
